import numpy as np
from matplotlib import pyplot as plt
import sys
import os
import platform


# make pyreco available locally
if platform.system() == 'Windows':  # WINDOWS
    curr_loc = os.getcwd()
    pyreco_path = os.path.join('C:\\',*curr_loc.split('\\')[1:-1], 'src')
    sys.path.append(pyreco_path)
elif platform.system() == 'Darwin':  # MAC
    curr_loc = os.getcwd()
    pyreco_path = curr_loc + '/src'
    sys.path.append(pyreco_path)
    
elif platform.system() == 'Linux':  # LINUX
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    src_path = os.path.join(project_root, 'src')
    sys.path.insert(0, src_path)

from pyreco.utils_data import sequence_to_sequence
from pyreco.custom_models import RC
from pyreco.layers import InputLayer, ReadoutLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.plotting import r2_scatter
from pyreco.network_prop_extractor import NetworkQuantifier

"""
Classic RC example
"""

# some testing data: predict a sine signal.
X_train, X_test, y_train, y_test = sequence_to_sequence(name='sine_pred', n_batch=10, n_states=2, n_time=1000)

# set the dimensions
input_shape = (X_train.shape[1], X_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])

num_init_nodes = 40
max_perf_drop = 0.5  # allow the network to reduce performance by x% w.r.t. to original one

model_rc = RC()
model_rc.add(InputLayer(input_shape=input_shape))
model_rc.add(RandomReservoirLayer(nodes=num_init_nodes, density=0.2, activation='tanh', leakage_rate=0.9,
                                  fraction_input=0.5))
model_rc.add(ReadoutLayer(output_shape, fraction_out=0.5))

# Compile the model
model_rc.compile(optimizer='ridge', metrics=['mean_squared_error'])

# Create a custom NetworkPropExtractor
prop_extractor = NetworkQuantifier(quantities=['density', 'spectral_radius', 'clustering_coefficient'])

# Train the model with pruning
history = model_rc.fit_prune(X_train, y_train, max_perf_drop=max_perf_drop, frac_rem_nodes=0.2, #patience=5,
                             prop_extractor=prop_extractor)

# Plot pruning results
plt.figure(figsize=(12, 8))
plt.axhline(y=history['pruned_nodes_scores'][0] * (1+max_perf_drop), linestyle='dashed', color='gray',
            label=f'threshold (+{max_perf_drop*100:.0f}%)')
plt.plot(history['num_nodes'], history['pruned_nodes_scores'], marker='.',
         markersize=8, label='pruned RC')
plt.plot(history['num_nodes'][0], history['pruned_nodes_scores'][0], color='red', marker='.', markersize=12, label='original RC')
plt.ylabel('loss')
plt.xlabel('number of nodes')
plt.title(f'initial reservoir size = {num_init_nodes}, final reservoir size = {history["num_nodes"][-1]}')
plt.legend()
# plt.savefig('pruning_results.png')
plt.show()

# Plot network properties
fig, axs = plt.subplots(len(history['network_properties'][0]), 1, figsize=(12, 4*len(history['network_properties'][0])))
fig.suptitle('Network Properties During Pruning')

for i, prop in enumerate(history['network_properties'][0].keys()):
    values = [props[prop] for props in history['network_properties']]
    axs[i].plot(history['num_nodes'], values, marker='o')
    axs[i].set_title(prop.capitalize())
    axs[i].set_xlabel('Number of Nodes')
    axs[i].set_ylabel('Value')
    axs[i].grid(True)

plt.tight_layout()
# plt.savefig('network_properties.png')
plt.show()

# Make predictions for new data using the pruned reservoir
y_pred = model_rc.predict(X_test)
print(f'shape of predictions on test set: {y_pred.shape}')

# Evaluate the model
loss_rc = model_rc.evaluate(X_test, y_test, metrics=['mae'])
print(f'Test model loss: {loss_rc}')

# plot predictions vs. ground truth
r2_scatter(y_true=y_test, y_pred=y_pred)

plt.figure()
plt.plot(y_test[0,:,0], label='ground truth', marker='.')
plt.plot(y_pred[0,:,0], label='prediction', marker='.')
plt.title('test set')
plt.legend()
plt.xlabel('time')
plt.ylabel('x1')
plt.show()

def print_and_compare_weights(model, history):
    """
    Prints and compares the reservoir and readout weights after pruning.
    
    Args:
    model (RC): The trained and pruned RC model.
    history (dict): The history dictionary returned by fit_prune.
    """
    print("\nComparing weights after pruning:")
    
    # Get the final weights
    final_reservoir_weights = model.reservoir_layer.weights
    final_readout_weights = model.readout_layer.weights
    
    # Get the original number of nodes
    original_nodes = history['num_nodes'][0]
    final_nodes = history['num_nodes'][-1]
    
    print(f"Original nodes: {original_nodes}, Final nodes: {final_nodes}")
    
    print("\nFinal Reservoir Weights (shape: {}):\n".format(final_reservoir_weights.shape))
    print(final_reservoir_weights)
    
    print("\nFinal Readout Weights (shape: {}):\n".format(final_readout_weights.shape))
    print(final_readout_weights)
    
    # Check consistency
    pruned_nodes = set(range(original_nodes)) - set(range(final_nodes))
    inconsistencies = []
    
    for node in pruned_nodes:
        if node < final_readout_weights.shape[0] and not np.allclose(final_readout_weights[node], 0):
            inconsistencies.append(node)
    
    if inconsistencies:
        print("\nInconsistencies found for nodes:", inconsistencies)
        print("These nodes were pruned from the reservoir but have non-zero readout weights.")
        print("\nInconsistent Readout Weights:")
        for node in inconsistencies:
            print(f"Node {node}: {final_readout_weights[node]}")
        
        print("\nCorresponding Reservoir Weights:")
        for node in inconsistencies:
            if node < final_reservoir_weights.shape[0]:
                print(f"Node {node} row: {final_reservoir_weights[node]}")
                print(f"Node {node} column: {final_reservoir_weights[:, node]}")
            else:
                print(f"Node {node} is outside the range of the reservoir weights matrix.")
    else:
        print("\nNo inconsistencies found. Pruning is correctly reflected in both layers.")

print_and_compare_weights(model_rc, history)







def RemoveTransients_Res(ResStates, Transients):
    return ResStates[Transients:,:]

def RemoveTransient_Inps(X, Transients):
    return X[:,Transients:,:]

def RemoveTransient_Outs(Y, Transients):
    return Y[:,Transients:,:]


def TransientRemover(What: str, ResStates, X, Y, Transients: int):

    if What=='RX':
        return RemoveTransients_Res(ResStates, Transients), RemoveTransient_Inps(X, Transients)
    if What == 'RXY':
        return RemoveTransients_Res(ResStates, Transients), RemoveTransient_Inps(X, Transients), RemoveTransient_Outs(Y, Transients)



import numpy as np
from abc import ABC
from typing import Union
import copy
import multiprocessing
from functools import partial

from pyreco.layers import (
    Layer,
    InputLayer,
    ReservoirLayer,
    ReadoutLayer,
)
from .optimizers import Optimizer, assign_optimizer
from .metrics import assign_metric
from .utils_networks import (
    gen_init_states,
    # set_spec_rad,
    # is_zero_col_and_row,
    # remove_node,
    get_num_nodes,
    # compute_spec_rad,
)
# from .metrics import assign_metric
from .network_prop_extractor import NetworkQuantifier  # NodePropExtractor


def sample_random_nodes(total_nodes: int, fraction: float):
    """
    Select a subset of randomly chosen nodes.

    Args:
    total_nodes (int): Total number of available nodes.
    fraction (float): Fraction of nodes to select.

    Returns:
    np.ndarray: Array of randomly selected node indices.
    """
    return np.random.choice(
        total_nodes, size=int(total_nodes * fraction), replace=False
    )


def discard_transients_indices(n_batches, n_timesteps, transients):
    indices_to_remove = []
    for i in range(n_batches * n_timesteps):
        t = i % n_timesteps  # Current timestep within the batch
        if t < transients:
            indices_to_remove.append(i)
    return indices_to_remove


class CustomModel(ABC):
    """
    Abstract base class for custom reservoir computing model.

    Has a syntax similar to the one of TensorFlow model API,
    e.g. using the model.add() statement to add layers to the model.

    A model hast an input layer, a reservoir layer and a readout layer.
    """

    def __init__(self):
        """
        Initialize the CustomModel with empty layers and default values.
        """
        # Initialize layers
        self.input_layer: InputLayer
        self.reservoir_layer: ReservoirLayer
        self.readout_layer: ReadoutLayer

        # Initialize hyperparameters
        self.metrics = []
        self.metrics_fun = []
        self.optimizer: Optimizer
        self.discard_transients = 0

        # Initialize other attributes
        self.num_trainable_weights: int

    def add(self, layer: Layer):
        """
        Add a layer to the model.

        Is type-sensitive and will assign the layer to the correct attribute.

        Args:
        layer (Layer): Layer to be added to the model.
        """
        if isinstance(layer, InputLayer):
            self.input_layer = layer
        elif issubclass(type(layer), ReservoirLayer):
            self.reservoir_layer = layer
        elif isinstance(layer, ReadoutLayer):
            self.readout_layer = layer

    # TODO: the following method should be implemented in the CustomModel class
    #   def _set_readin_nodes(self, nodes: Union[list, np.ndarray] = None):

    """
    The setter methods are used to set the parameters of the model.
    """

    def _set_readout_nodes(self, nodes: Union[list, np.ndarray] = None):
        """
        Set the nodes that will be linked to the output.

        Args:
        nodes (Union[list, np.ndarray], optional): Specific nodes to use for readout
        provided as indices. If None, randomly sample nodes.
        """
        if nodes is None:
            nodes = sample_random_nodes(
                total_nodes=self.reservoir_layer.nodes,
                fraction=self.readout_layer.fraction_out,
            )
        self.readout_layer.readout_nodes = nodes

    def _set_optimizer(self, optimizer: Union[str, Optimizer]):
        """
        Set the optimizer that will find the readout weights.

        Args:
        optimizer (Union[str, Optimizer]): Name of the optimizer or an Optimizer
        instance.
        """
        self.optimizer = assign_optimizer(optimizer)

    def _set_metrics(self, metrics: Union[list, str]):
        """
        Set the metric(s) for model evaluation.

        Args:
        metrics (Union[list, str]): List of metric names or a single metric name.
        """
        if isinstance(metrics, str):  # only single metric given
            self.metrics = [metrics]
        else:
            self.metrics = metrics  # if metrics is a list of strings.

        # assign the metric functions (callable) according to the metric names
        self.metrics_fun = []  # has been initialized, we reset it here
        for metric in self.metrics:
            self.metrics_fun.append(assign_metric(metric))

    def _set_init_states(self, init_states=None, method=None):
        """
        Set the initial states of the reservoir nodes.

        Args:
        init_states (np.ndarray, optional): Array of initial states. If None, sample
        initial states using the specified method.
        method (str, optional): Method for sampling initial states.
        """
        if init_states is not None:
            if init_states.shape[0] != self.reservoir_layer.nodes:
                raise (
                    ValueError(
                        "initial states not matching the number of reservoir nodes!"
                    )
                )
            self.reservoir_layer.set_initial_state(r_init=init_states)
        elif (init_states is None) and (method is not None):
            init_states = gen_init_states(self.reservoir_layer.nodes, method)
            self.reservoir_layer.set_initial_state(r_init=init_states)
        else:
            raise (
                ValueError(
                    "provide either an array of initial states or a method for sampling"
                )
            )

    def _connect_input_to_reservoir(self, nodes: Union[list, np.ndarray] = None):
        """
        Wire input layer with reservoir layer. Creates a random matrix of shape
        [nodes x n_states], i.e. number of reservoir nodes x state dimension of input.
        If no full connection is desired, a fraction of nodes will be selected according
        to the fraction_input parameter of the reservoir layer.

        Args:
        nodes (Union[list, np.ndarray], optional): Specific nodes to connect to the
        input. If None, randomly sample nodes.
        """
        num_input_states = self.input_layer.n_states
        num_reservoir_nodes = self.reservoir_layer.nodes

        # generate random input connection matrix [nodes, n_states]
        full_input_weights = np.random.randn(num_input_states, num_reservoir_nodes)

        # select read-in node indices according to the fraction specified by the user
        if nodes is None:
            input_nodes = sample_random_nodes(
                total_nodes=num_reservoir_nodes,
                fraction=self.reservoir_layer.fraction_input,
            )
        else:
            input_nodes = nodes

        # mask the input weights matrix to only have the selected nodes
        mask = np.zeros_like(full_input_weights) 
        mask[:, input_nodes] = 1
        self.input_layer.weights = full_input_weights * mask

    def compile(
        self,
        optimizer: str = "ridge",
        metrics: list = ["mse"],
        discard_transients: int = 0,
    ):
        """
        Configure the model for training.

        Args:
        optimizer (str): Name of the optimizer.
        metrics (list): List of metric names.
        discard_transients (int): Number of initial transient timesteps to discard.
        """
        # set the metrics (like in TensorFlow)
        self._set_metrics(metrics)

        # set the optimizer that will find the readout weights
        self._set_optimizer(optimizer)

        # 1. check consistency of layers, data shapes etc.
        # TODO: do we have input, reservoir and readout layer?
        # TODO: are all shapes correct on input and output side?

        # 2. Sample the input connections: create W_in read-in weight matrix
        self._connect_input_to_reservoir()  # check for dependency injection here!

        # 3. Select readout nodes according to the fraction specified by the user in the readout layer
        self._set_readout_nodes()

        # 4. set reservoir initialization

        # 5. discarding transients from reservoir states
        if discard_transients < 0:
            raise (ValueError("discard_transients must be >= 0!"))
        self.discard_transients = int(
            discard_transients
        )  # will not remove transients if 0

    def compute_reservoir_state(self, X: np.ndarray, seed=None) -> np.ndarray:
        """
        Vectorized computation of reservoir states with batch processing.

        Args:
            X (np.ndarray): Input data of shape [n_batch, n_timesteps, n_states]
            seed (int, optional): Random seed for reproducibility

        Returns:
            np.ndarray: Reservoir states of shape [(n_batch * n_timesteps), N]
        """
        # Extract shapes and parameters
        n_batch, n_time, n_states = X.shape
        N = self.reservoir_layer.nodes
        activation_name = self.reservoir_layer.activation
        alpha = self.reservoir_layer.leakage_rate
        A = self.reservoir_layer.weights
        W_in = self.input_layer.weights

        # Pre-allocate arrays
        states = np.zeros((n_batch, n_time + 1, N))
        states[:, 0] = self.reservoir_layer.initial_res_states

        # vectorized computation of reservoir states
        input_contrib = np.einsum("ij,btj->bti", W_in.T, X)
        for t in range(n_time):
            reservoir_contrib = np.einsum("ij,bj->bi", A, states[:, t])
            states[:, t + 1] = (1 - alpha) * states[
                :, t
            ] + alpha * self.reservoir_layer.activation_fun(
                reservoir_contrib + input_contrib[:, t]
            )

        return states[:, 1:].reshape(-1, N)

    def fit(
        self, X: np.ndarray, y: np.ndarray, n_init: int = 1, store_states: bool = False
    ):
        """
        Optimized training with batch processing.
        """
        n_batch, n_time, n_states_out = X.shape[0], X.shape[1], y.shape[-1]
        n_nodes = self.reservoir_layer.nodes

        # Pre-allocate arrays for storing results
        n_R0 = np.zeros((n_init, n_nodes))
        n_weights = np.zeros((n_init, n_nodes, n_states_out))
        n_scores = np.zeros(n_init)
        n_res_states = [] if store_states else None

        # Get metric functions that scores the model performance
        metric_fun = (
            self.metrics_fun[0]
            if self.metrics_fun
            else assign_metric("mean_squared_error")
        )

        # Batch process multiple initializations
        for i in range(n_init):
            if n_init > 1:
                print(f"initialization {i}/{n_init}: computing reservoir states")

            # Set initial states
            self._set_init_states(method=self.reservoir_layer.init_res_sampling)
            n_R0[i] = self.reservoir_layer.initial_res_states

            # Compute reservoir states
            reservoir_states = self.compute_reservoir_state(X)

            # Handle transients efficiently
            if self.discard_transients > 0:
                indices_to_remove = discard_transients_indices(
                    n_batch, n_time, self.discard_transients
                )  # by juan
                reservoir_states = np.delete(
                    reservoir_states, indices_to_remove, axis=0
                )
                # now the array should have the size of (n_batch*(n_time-discard), n_nodes)

                # remove the transients from the targets
                y = y[:, self.discard_transients :, :]

                # update the value of n_time
                n_time -= self.discard_transients

            # Efficient masking for readout nodes
            A = np.zeros_like(reservoir_states)
            A[:, self.readout_layer.readout_nodes] = reservoir_states[
                :, self.readout_layer.readout_nodes
            ]

            # Reshape targets efficiently
            b = y.reshape(n_batch * n_time, n_states_out)

            # Solve regression problem y = W_out * R
            if n_batch == 1:
                self.readout_layer.weights = np.expand_dims(
                    self.optimizer.solve(A=A, b=b), axis=-1
                )
            else:
                self.readout_layer.weights = self.optimizer.solve(A=A, b=b)

            # is there is only a single system state to predict, we need to add that dim
            # TODO: move this to the sanity checks and add an artificial dimension prior to fitting!
            if self.readout_layer.weights.ndim == 1:
                self.readout_layer.weights = np.expand_dims(self.readout_layer.weights, axis=-1)

            # store weights for this initialization
            n_weights[i] = self.readout_layer.weights

            # Compute score
            n_scores[i] = metric_fun(y, self.predict(X=X))

            if store_states:
                n_res_states.append(reservoir_states)

        # Select best initialization
        idx_optimal = np.argmin(n_scores)
        self.reservoir_layer.set_initial_state(n_R0[idx_optimal])
        self.readout_layer.weights = n_weights[idx_optimal]

        # Update trainable weights count
        self.num_trainable_weights = self.reservoir_layer.weights.size

        # Build history dictionary
        history = {
            "init_res_states": n_R0,
            "readout_weights": n_weights,
            "train_scores": n_scores,
        }

        if store_states:
            history["res_states"] = n_res_states

        return history

    def fit_evolve(self, X: np.ndarray, y: np.ndarray):
        # build an evolving reservoir computer: performance-dependent node addition and removal

        history = None
        return history

    def evaluate_node_removal(
        self, X, y, loss_fun, init_score, del_idx, current_num_nodes
    ):
        # Create a deep copy of the current model
        temp_model = copy.deepcopy(self)

        # Remove node from the temporary model
        temp_model.remove_node(del_idx)

        # Train the temporary model
        temp_model.fit(X, y)
        y_discarded = y[:, self.discard_transients :, :]
        # Evaluate the temporary model
        temp_score = loss_fun(y_discarded, temp_model.predict(X=X))

        print(
            f"Pruning node {del_idx} / {current_num_nodes}: loss = {temp_score:.5f}, original loss = {init_score:.5f}"
        )

        return temp_score

    def fit_prune(
        self,
        X: np.ndarray,
        y: np.ndarray,
        loss_metric="mse",
        max_perf_drop=0.1,
        frac_rem_nodes=0.20,
        patience=None,
        prop_extractor=None,
    ):
        """
        Build a reservoir computer by performance-informed pruning of the initial reservoir network.

        This method prunes the network down to better performance OR a tolerated performance reduction.

        Args:
            X (np.ndarray): Input data of shape [n_batch, n_time_in, n_states_in]
            y (np.ndarray): Target data of shape [n_batch, n_time_out, n_states_out]
            loss_metric (str): Metric for performance-informed node removal. Must be a member of existing metrics in pyReCo.
            max_perf_drop (float): Maximum allowed performance drop before stopping pruning. Default: 0.1 (10%)
            frac_rem_nodes (float): Fraction of nodes to attempt to remove in each iteration. Default: 0.01 (1%)
            patience (int): Number of consecutive performance decreases allowed before early stopping
            prop_extractor (object): Object to extract network properties during pruning

        Returns:
            dict: History of the pruning process
        """

        # Ensure frac_rem_nodes is within the valid range [0, 1]
        frac_rem_nodes = max(0.0, min(1.0, frac_rem_nodes))

        # Get a callable loss function for performance-informed node removal
        loss_fun = assign_metric(loss_metric)

        # Initialize reservoir states
        self._set_init_states(method=self.reservoir_layer.init_res_sampling)

        # Get size of original reservoir
        num_nodes = self.reservoir_layer.weights.shape[0]

        # Set default patience if not specified
        if patience is None:
            patience = num_nodes

        # Compute initial score of full network on training set
        self.fit(X, y)
        y_discarded = y[:, self.discard_transients :, :]
        init_score = loss_fun(y_discarded, self.predict(X=X))

        def keep_pruning(init_score, current_score, max_perf_drop):
            """
            Determine if pruning should continue based on current performance.
            """
            if current_score < (init_score * (1.0 + max_perf_drop)):
                return True
            else:
                print("Pruning stopping criterion reached.")
                return False

        # Initialize property extractor if not provided. TODO needs to be improved
        if prop_extractor is None:
            prop_extractor = NetworkQuantifier()

        # Initialize pruning variables
        i = 0
        current_score = init_score
        current_num_nodes = get_num_nodes(self.reservoir_layer.weights)
        score_per_node = []
        history = {
            "pruned_nodes": [-1],
            "pruned_nodes_scores": [init_score],
            "num_nodes": [current_num_nodes],
            "network_properties": [],
        }

        # Extract initial network properties
        initial_props = prop_extractor.extract_properties(self.reservoir_layer.weights)
        history["network_properties"].append(initial_props)

        consecutive_increases = 0
        best_score = init_score

        best_model = copy.deepcopy(self)

        # Main pruning loop
        while i < num_nodes:
            print(f"Pruning iteration {i}")

            # Calculate number of nodes to try removing this iteration
            num_nodes_to_try = max(1, int(current_num_nodes * frac_rem_nodes))

            score_per_node.append([])
            max_loss = init_score

            # Prepare the partial function for multiprocessing
            evaluate_func = partial(
                self.evaluate_node_removal,
                X,
                y,
                loss_fun,
                init_score,
                current_num_nodes=current_num_nodes,
            )

            # Use multiprocessing to evaluate node removals in parallel
            with multiprocessing.Pool() as pool:
                results = pool.map(evaluate_func, range(current_num_nodes))

            # Process the results
            score_per_node[i] = results
            max_loss = max(max_loss, max(results))

            # Find nodes which affect the loss the least
            max_loss = max_loss + 1
            score_per_node[i] = [
                max_loss if x is None else x for x in score_per_node[i]
            ]
            sorted_indices = np.argsort(score_per_node[i])
            nodes_to_remove = sorted_indices[:num_nodes_to_try]

            if keep_pruning(init_score, current_score, max_perf_drop):
                # Remove node from all layers
                self.remove_node(nodes_to_remove)

                # Retrain and evaluate
                self.fit(X, y)
                y_discarded = y[:, self.discard_transients :, :]
                current_score = loss_fun(y_discarded, self.predict(X=X))
                rel_score = (current_score - init_score) / init_score * 100

                current_num_nodes = self.reservoir_layer.nodes

                print(
                    f"Removing node {nodes_to_remove}: new loss = {current_score:.5f}, original loss = {init_score:.5f} ({rel_score:+.2f} %); {current_num_nodes} nodes remain"
                )

                # Check for early stopping and update best model
                if current_score < best_score:
                    best_score = current_score
                    best_model = copy.deepcopy(self)
                    consecutive_increases = 0
                else:
                    consecutive_increases += 1
                    if consecutive_increases >= patience:
                        print(
                            f"Stopping pruning: Loss increased for {patience} consecutive iterations."
                        )
                        break

                # Extract and store network properties
                network_props = prop_extractor.extract_properties(
                    self.reservoir_layer.weights
                )
                history["network_properties"].append(network_props)

                # Update pruning history
                history["pruned_nodes"].append(nodes_to_remove.tolist())
                history["pruned_nodes_scores"].append(
                    [score_per_node[i][node] for node in nodes_to_remove]
                )
                history["num_nodes"].append(current_num_nodes)
            else:
                break

            i += 1

        # Add best score to history
        history["best_score"] = best_score

        return history, best_model

    # @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Make predictions for given input (single-step prediction).

        Args:
        X (np.ndarray): Input data of shape [n_batch, n_timestep, n_states]
        one_shot (bool): If True, don't re-initialize reservoir between samples.

        Returns:
        np.ndarray: Predictions of shape [n_batch, n_timestep, n_states]
        """
        # makes prediction for given input (single-step prediction)
        # expects inputs of shape [n_batch, n_timestep, n_states]
        # returns predictions in shape of [n_batch, n_timestep, n_states]

        # one_shot = True will *not* re-initialize the reservoir from sample to sample. Introduces a dependency on the
        # sequence by which the samples are given

        # TODO Merten: return some random number that have the correct shape

        # TODO: external function that is going to check the dimensionality
        # and raise an error if shape is not correct
        n_batch, n_time, n_states = X.shape[0], X.shape[1], X.shape[2]
        n_nodes = self.reservoir_layer.nodes

        # iterate over batch to obtain predictions
        reservoir_states = self.compute_reservoir_state(X)

        # Removing transients AKA Warm-up and update time
        # TODO: this is a lot of boilerplate code. @Juan reuse the function from .fit
        if self.discard_transients >= n_time:
            raise ValueError(
                f"Cannot discard {self.discard_transients} as the number of time steps is {n_time}"
            )
        if self.discard_transients > 0:

            # removes the first <discard_transients> from the reservoir states and from the targets
            # reservoir_states.shape is 2d, as we concatenated along the batch dimension: [n_time * n_batch, n_nodes]
            # hence we have to remove slices from the state matrix, or re-shape it into 3D, cut off some time steps
            # for each batch, and then reshape to 2D again.
            # TODO: please check if the reshaping really is correct, i.e. such that the first n_time entries of reservoir_states are the continuous reservoir states!
            indices_to_remove = discard_transients_indices(
                n_batch, n_time, self.discard_transients
            )
            reservoir_states = np.delete(reservoir_states, indices_to_remove, axis=0)
            # now the array should have the size of (n_batch*(n_time-discard), n_nodes)

            # update the value of n_time
            n_time = -self.discard_transients

            # reservoir_states, X, y = TransientRemover('RXY', reservoir_states, X, y, self.discard_transients)

        # make predictions y = R * W_out, W_out has a shape of [n_out, N]
        y_pred = np.dot(reservoir_states, self.readout_layer.weights)

        # reshape predictions into 3D [n_batch, n_time_out, n_state_out]
        n_time_out = int(y_pred.shape[0] / n_batch)
        n_states_out = y_pred.shape[-1]
        y_pred = y_pred.reshape(n_batch, n_time_out, n_states_out)

        return y_pred

    # @abstractmethod
    def evaluate(
        self, X: np.ndarray, y: np.ndarray, metrics: Union[str, list, None] = None
    ) -> tuple:
        """
        Evaluate metrics on predictions made for input data.

        Args:
        X (np.ndarray): Input data of shape [n_batch, n_timesteps, n_states]
        y (np.ndarray): Target data of shape [n_batch, n_timesteps_out, n_states_out]
        metrics (Union[str, list, None], optional): List of metric names or a single metric name. If None, use metrics from .compile()

        Returns:
        tuple: Metric values
        """
        # evaluate metrics on predictions made for input data
        # expects: X of shape [n_batch, n_timesteps, n_states]
        # expects: y of shape [n_batch, n_timesteps_out, n_states_out]
        # depends on self.metrics = metrics from .compile()
        # returns float, if multiple metrics, then in given order (TODO: implement this)

        if (
            metrics is None
        ):  # user did not specify metric, take the one(s) given to .compile()
            metrics = self.metrics
        if type(metrics) is str:  # make sure that we are working with lists of strings
            metrics = [metrics]

        # self.metrics_available = ['mse', 'mae        #
        # eval_metrics = self.metrics + metrics  # combine from .compile and user specified
        # eval_metrics = list(set(eval_metrics))  # removes potential duplicates

        # get metric function handle from the list of metrics specified as str
        metric_funs = [assign_metric(m) for m in metrics]

        # make predictions
        y_pred = self.predict(X)

        # remove some initial transients from the ground truth if discard transients is active
        n_time = y.shape[1]
        if self.discard_transients >= n_time:
            raise ValueError(
                f"Cannot discard {self.discard_transients} as the number of time steps is {n_time}"
            )

        if self.discard_transients > 0:
            y = y[:, self.discard_transients :, :]

        # get metric values
        metric_values = []
        for _metric_fun in metric_funs:
            metric_values.append(float(_metric_fun(y, y_pred)))

        return metric_values

    # @abstractmethod
    def get_params(self, deep=True):
        """
        Get parameters for scikit-learn compatibility.

        Args:
        deep (bool): If True, return a deep copy of parameters.

        Returns:
        dict: Dictionary of model parameters.
        """
        # needed for scikit-learn compatibility
        return {
            "input_layer": self.input_layer,
            "reservoir_layer": self.reservoir_layer,
            "readout_layer": self.readout_layer,
        }

    # @abstractmethod
    def save(self, path: str):
        """
        Store the model to disk.

        Args:
        path (str): Path to save the model.
        """
        # store the model to disk
        pass

    def plot(self, path: str):
        """
        Print the model to some figure file.

        Args:
        path (str): Path to save the figure.
        """
        # print the model to some figure file
        pass

    def remove_node(self, node_indices):
        """
        Remove one or multiple nodes from all relevant layers of the reservoir computer.

        Args:
        node_indices (int or list or np.array): Index or indices of the nodes to be removed.
        """
        # Convert single integer to list
        if isinstance(node_indices, int):
            node_indices = [node_indices]

        # Remove nodes from reservoir layer weights
        self.reservoir_layer.weights = np.delete(
            self.reservoir_layer.weights, node_indices, axis=0
        )
        self.reservoir_layer.weights = np.delete(
            self.reservoir_layer.weights, node_indices, axis=1
        )

        # Remove nodes from initial reservoir states
        self.reservoir_layer.initial_res_states = np.delete(
            self.reservoir_layer.initial_res_states, node_indices, axis=0
        )

        # Remove nodes from input layer weights
        self.input_layer.weights = np.delete(
            self.input_layer.weights, node_indices, axis=1
        )

        # Remove nodes from readout layer weights
        self.readout_layer.weights = np.delete(
            self.readout_layer.weights, node_indices, axis=0
        )

        # Update readout nodes
        mask = np.ones(len(self.readout_layer.readout_nodes), dtype=bool)
        for idx in node_indices:
            mask[self.readout_layer.readout_nodes == idx] = False
        self.readout_layer.readout_nodes = self.readout_layer.readout_nodes[mask]

        # Adjust the indices of the remaining readout nodes
        for idx in sorted(node_indices, reverse=True):
            self.readout_layer.readout_nodes[
                self.readout_layer.readout_nodes > idx
            ] -= 1

        # Update node count
        self.reservoir_layer.nodes -= len(node_indices)


class RC(CustomModel):  # the non-auto version
    """
    Non-autonomous version of the reservoir computer.
    """

    def __init__(self):
        # at the moment we do not have any arguments to pass
        super().__init__()


class AutoRC(CustomModel):
    """
    Autonomous version of the reservoir computer.
    """

    def __init__(self):
        pass

    def predict_ar(self, X: np.ndarray, n_steps: int = 10):
        """
        Perform auto-regressive prediction (time series forecasting).

        Args:
        X (np.ndarray): Initial input data.
        n_steps (int): Number of steps to predict into the future.

        Returns:
        np.ndarray: Predicted future states.
        """
        pass


class HybridRC(CustomModel):
    """
    Hybrid version of the reservoir computer.
    """

    def __init__(self):
        pass


if __name__ == "__main__":
    print("hello")



import numpy as np
from matplotlib import pyplot as plt
from pyreco.models import ReservoirComputer as RC
from pyreco.plotting import r2_scatter

# generate some training data: map a sine to a cosine (learn a phase shift) with signal
# amplification (learn to increase the amplitude in the targets)

# generate 3 cycles of a sine (input) and of a cosine (output)
omega = np.pi
t = np.linspace(start=0, stop=3 * (2 * np.pi / omega), num=300, endpoint=True)
x = np.sin(omega * t)
y = 2 * np.cos(omega * t)

x_train = np.expand_dims(x, axis=(0, 2))  # obtain shape of [n_batch, n_time, n_states]
y_train = np.expand_dims(y, axis=(0, 2))

# fit a reservoir computer with 200 nodes and make predictions on the training set
model = RC(num_nodes=200, activation="tanh")
model.fit(x_train, y_train)
y_pred = model.predict(x_train)
print(f"shape of predicted array: {y_pred.shape}")

# evaluate some metrics (for simplicity on the train set)
metric_value = model.evaluate(X=x_train, y=y_train, metrics=["mse", "mae"])
print(f"scores:{metric_value}")

plt.figure(figsize=(10, 4), dpi=100)
plt.plot(y_train[0, :, 0], label="ground truth", marker=".", color="#1D3557")
plt.plot(y_pred[0, :, 0], label="prediction", marker=".", color="#E63946")
plt.legend()
plt.xlabel("time")
plt.tight_layout()
# plt.savefig('model_predictions.png')
plt.show()

# plot an R^2 - like graphic
r2_scatter(y_train, y_pred)




import sys
import os
import platform
import numpy as np
import time
from matplotlib import pyplot as plt
import copy
import threading

# Platform-specific path setup
if platform.system() == 'Linux':
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    src_path = os.path.join(project_root, 'src')
    sys.path.insert(0, src_path)

# Import RC implementations and components
import pyreco.custom_models_1 as rc_old
import pyreco.custom_models as rc_new
from pyreco.layers import InputLayer, ReadoutLayer, RandomReservoirLayer
from pyreco.utils_data import sequence_to_sequence
from rc_profiler import RCProfiler

def setup_model(model_class, input_shape, output_shape):
    """Setup RC model with consistent parameters."""
    model = model_class.RC()
    model.add(InputLayer(input_shape=input_shape))
    model.add(RandomReservoirLayer(
        nodes=100,
        density=0.1,
        activation='sigmoid',
        leakage_rate=0.1,
        fraction_input=0.5
    ))
    model.add(ReadoutLayer(output_shape, fraction_out=0.99))
    model.compile(optimizer='ridge', metrics=['mse'])
    return model

def run_training_benchmark(model, X, y, n_init=1):
    """Run training benchmark and collect metrics."""
    profiler = RCProfiler(model)
    
    # Start monitoring
    monitor = threading.Thread(target=profiler.start_monitoring)
    monitor.daemon = True
    monitor.start()
    
    try:
        start_time = time.time()
        start_mem = profiler.process.memory_info().rss / 1024 / 1024
        
        # Train model
        with profiler.profile_section("Model Training"):
            history = model.fit(X, y, n_init=n_init)
            mse = model.evaluate(X, y, metrics=['mse'])[0]
        
        metrics = {
            'time': time.time() - start_time,
            'memory': profiler.process.memory_info().rss / 1024 / 1024 - start_mem,
            'mse': mse,
            'cpu': np.mean(profiler.metrics['cpu'])
        }
        
    finally:
        profiler.stop_monitoring()
        monitor.join(timeout=1.0)
    
    return metrics

def run_prediction_benchmark(model, X, y):
    """Run prediction benchmark and collect metrics."""
    profiler = RCProfiler(model)
    
    # Start monitoring
    monitor = threading.Thread(target=profiler.start_monitoring)
    monitor.daemon = True
    monitor.start()
    
    try:
        start_time = time.time()
        start_mem = profiler.process.memory_info().rss / 1024 / 1024
        
        # Make predictions
        with profiler.profile_section("Model Prediction"):
            y_pred = model.predict(X)
            mse = model.evaluate(X, y, metrics=['mse'])[0]
        
        metrics = {
            'time': time.time() - start_time,
            'memory': profiler.process.memory_info().rss / 1024 / 1024 - start_mem,
            'mse': mse,
            'cpu': np.mean(profiler.metrics['cpu'])
        }
        
    finally:
        profiler.stop_monitoring()
        monitor.join(timeout=1.0)
    
    return metrics

def run_pruning_benchmark(model, X, y):
    """Run pruning benchmark and collect metrics."""
    profiler = RCProfiler(model)
    
    # Start monitoring
    monitor = threading.Thread(target=profiler.start_monitoring)
    monitor.daemon = True
    monitor.start()
    
    try:
        start_time = time.time()
        start_mem = profiler.process.memory_info().rss / 1024 / 1024
        
        # Run pruning
        with profiler.profile_section("Model Pruning"):
            if isinstance(model, rc_old.RC):
                history = model.fit_prune(X, y, loss_metric='mse', max_perf_drop=0.1)
            else:  # New implementation
                history, best_model = model.fit_prune(X, y, loss_metric='mse', max_perf_drop=0.1)
        
        # Get final MSE
        mse = model.evaluate(X, y, metrics=['mse'])[0]
        
        metrics = {
            'time': time.time() - start_time,
            'memory': profiler.process.memory_info().rss / 1024 / 1024 - start_mem,
            'mse': mse,
            'cpu': np.mean(profiler.metrics['cpu']),
            'history': history,
            'nodes_removed': history['num_nodes'][0] - history['num_nodes'][-1] if 'num_nodes' in history else None,
            'mse_progression': history['pruned_nodes_scores'] if 'pruned_nodes_scores' in history else None
        }
        
    finally:
        profiler.stop_monitoring()
        monitor.join(timeout=1.0)
    
    return metrics

def run_all_benchmarks(num_runs=5):
    """Run all benchmarks and collect results."""
    # Generate test data
    X_train, X_test, y_train, y_test = sequence_to_sequence(
        name='sincos2', n_batch=10, n_states=1, n_time=200)
    
    input_shape = (X_train.shape[1], X_train.shape[2])
    output_shape = (y_train.shape[1], y_train.shape[2])
    
    results = {
        'single_init': {'old': [], 'new': []},
        'multi_init': {'old': [], 'new': []},
        'prediction': {'old': [], 'new': []},
        'pruning': {'old': [], 'new': []}
    }
    
    # Run benchmarks
    for run in range(num_runs):
        print(f"\nRun {run + 1}/{num_runs}")
        
        for model_type in ['old', 'new']:
            model_class = rc_old if model_type == 'old' else rc_new
            
            # Single init training
            model = setup_model(model_class, input_shape, output_shape)
            results['single_init'][model_type].append(
                run_training_benchmark(model, X_train, y_train, n_init=1))
            
            # Multi init training
            model = setup_model(model_class, input_shape, output_shape)
            results['multi_init'][model_type].append(
                run_training_benchmark(model, X_train, y_train, n_init=5))
            
            # Prediction
            model = setup_model(model_class, input_shape, output_shape)
            model.fit(X_train, y_train)  # Train first
            results['prediction'][model_type].append(
                run_prediction_benchmark(model, X_test, y_test))
            
            # Pruning
            model = setup_model(model_class, input_shape, output_shape)
            results['pruning'][model_type].append(
                run_pruning_benchmark(model, X_train, y_train))
    
    return results

def plot_benchmark_results(results, save_path_prefix=None):
    """Create plots for each operation type."""
    operations = {
        'single_init': 'Training (Single Init)',
        'multi_init': 'Training (Multiple Init)',
        'prediction': 'Prediction',
        'pruning': 'Pruning'
    }
    
    metrics = ['time', 'memory', 'cpu', 'mse']
    metric_labels = ['Time (s)', 'Memory (MB)', 'CPU Usage (%)', 'MSE']
    
    for op_name, op_title in operations.items():
        fig, axes = plt.subplots(2, 2, figsize=(15, 10))
        fig.suptitle(f'{op_title} Performance Comparison')
        
        for idx, (metric, label) in enumerate(zip(metrics, metric_labels)):
            ax = axes[idx // 2, idx % 2]
            
            old_data = [r[metric] for r in results[op_name]['old']]
            new_data = [r[metric] for r in results[op_name]['new']]
            
            ax.boxplot([old_data, new_data], labels=['Old', 'New'])
            ax.set_title(label)
            ax.grid(True)
            
            # Calculate improvement
            improvement = ((np.mean(old_data) - np.mean(new_data)) / np.mean(old_data) * 100)
            ax.text(0.5, 0.95, f'Improvement: {improvement:+.1f}%', 
                   transform=ax.transAxes, ha='center')
        
        plt.tight_layout()
        
        if save_path_prefix:
            plt.savefig(f'{save_path_prefix}_{op_name}.png')
        plt.show()

if __name__ == "__main__":
    np.random.seed(42)  # For reproducibility
    
    print("Starting benchmarks...")
    results = run_all_benchmarks(num_runs=5)
    
    print("\nGenerating plots...")
    plot_benchmark_results(results, save_path_prefix='rc_benchmark')
    
    print("Benchmarking completed!")



"""
Minimal testing example for predicting a sine function (discrete time setting)

x(t) -> x(t+dt)

"""

import numpy as np
from matplotlib import pyplot as plt

"""
generate some training and testing data from the sine function
"""

omega = np.pi
t = np.linspace(start=0, stop=5 * omega, num=51, endpoint=True)
signal = np.sin(omega * t)

# convert into inputs (function at current time point) and outputs (function at next time point)
x = signal[:-1]
y = signal[1:]

# train-test split 80%
ratio = 0.8
n = len(x)
shuffle_idx = np.random.choice(n, size=n, replace=False)
train_idx, test_idx = shuffle_idx[:int(n * ratio)], shuffle_idx[int(n * ratio):]

# split data and create data shape [n_batch, n_features]
X_train, X_test = np.expand_dims(x[train_idx], axis=1), np.expand_dims(x[test_idx], axis=1)
y_train, y_test = np.expand_dims(y[train_idx], axis=1), np.expand_dims(y[test_idx], axis=1)
print(f'shape of training data: {X_train.shape}, shape of test data: {y_test.shape}')

"""
Modeling using feed-forward neural nets in TensorFlow (sequential API)
"""

import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Input

model_ann = Sequential()                                # Instantiate Sequential model
model_ann.add(Input(shape=(1,)))                        # Add input layer
model_ann.add(Dense(units=100, activation='sigmoid'))   # Add hidden Dense layer
model_ann.add(Dense(units=1, activation='linear'))      # Add output layer

model_ann.compile(optimizer='adam', loss='mean_squared_error')  # Compile the model
hist = model_ann.fit(X_train, y_train, epochs=500)              # Train the model for 1000 epochs
loss_ann = model_ann.evaluate(X_test, y_test)                       # Evaluate the model
print(f'Test model loss: {loss_ann}')
y_pred_ann = model_ann.predict(X_test)                          # Make predictions for new data


"""
Modeling using reservoir computing built on the pyreco library (ours)
"""
import sys
import os
import platform

# make pyReCo available locally
if platform.system() == 'Windows':  # WINDOWS
    curr_loc = os.getcwd()
    pyreco_path = os.path.join('C:\\',*curr_loc.split('\\')[1:-1], 'src')
    sys.path.append(pyreco_path)
elif platform.system() == 'Darwin':  # MAC
    curr_loc = os.getcwd()
    pyreco_path = curr_loc + '/src'
    sys.path.append(pyreco_path)


from pyreco.models import RC
from pyreco.layers import InputLayer, RandomReservoirLayer, ReadoutLayer

model_rc = RC()
model_rc.add(InputLayer(input_shape=(1,1)))
model_rc.add(RandomReservoirLayer(nodes=100, density=0.1, activation='sigmoid', leakage_rate=0.2, fraction_input=0.5))
model_rc.add(ReadoutLayer(output_shape=(1,1), fraction_out=0.6))

# model_rc.compile(optimizer='ridge', metrics=['mean_squared_error'])  # Compile the model
model_rc.fit(X_train, y_train)#, epochs=500)              # Train the model for 1000 epochs
y_pred_rc = model_rc.predict(X_test)                          # Make predictions for new data

loss_rc = model_rc.evaluate(X_test, y_test)                       # Evaluate the model
print(f'Test model loss: {loss_rc}')

"""
Evaluate prediction results
"""

fig = plt.figure()
plt.plot([np.min(y_test), np.max(y_test)], [np.min(y_test), np.max(y_test)], color='gray', linestyle='dashed')
plt.plot(y_test, y_pred_ann, linestyle='none', marker='.', color='blue', label='feed-forward ANN')
# plt.plot(y_test, y_pred_rc, linestyle='none', marker='.', color='red', label='RC')
plt.xlabel('true')
plt.ylabel('predicted')
plt.legend()
plt.tight_layout()
plt.show()



"""
Helper routines for networks
"""
import networkx as nx
import numpy as np


def gen_ER_graph(nodes: int, density: float, spec_rad: float = 0.9, directed: bool=True, seed=None):
    """Generate an Erdős-Rényi random graph with specified properties.
    
    Bug Fix Documentation:
    ---------------------
    Previous Bug:
        The line `G.remove_nodes_from(list(nx.isolates(G)))` removed isolated nodes from the graph
        before converting to a numpy array. This caused the final matrix to sometimes be smaller
        than the specified size (e.g., 29x29 instead of 30x30) when isolated nodes were present.
        This led to dimension mismatches in reservoir computations where other matrices expected
        the full size.

    Solution:
        Instead of removing isolated nodes, we now connect them to maintain the specified
        network size. This ensures consistency between the reservoir weight matrix and
        other matrices in the computation.

    Parameters:
        nodes (int): Number of nodes in the graph
        density (float): Desired connection density (0 to 1)
        spec_rad (float): Desired spectral radius
        directed (bool): Whether to create a directed graph
        seed (int, optional): Random seed for reproducibility

    Returns:
        np.ndarray: Adjacency matrix with shape (nodes, nodes)
    """
    # use networkx to generate a random graph
    G = nx.erdos_renyi_graph(nodes, density, seed=seed, directed=directed)

    # Instead of removing isolated nodes (old buggy behavior):
    # G.remove_nodes_from(list(nx.isolates(G)))
    
    # New: Connect isolated nodes to maintain matrix size
    isolated_nodes = list(nx.isolates(G))
    if isolated_nodes:
        non_isolated = list(set(G.nodes()) - set(isolated_nodes))
        for node in isolated_nodes:
            if non_isolated:
                target = np.random.choice(non_isolated)
                G.add_edge(node, target)
                if directed:
                    G.add_edge(target, node)

    GNet = nx.to_numpy_array(G)
    curr_spec_rad = max(abs(np.linalg.eigvals(GNet)))
    graph = GNet * spec_rad/curr_spec_rad

    return graph


def compute_density(network: np.ndarray) -> float:
    # compute density of a given adjacency matrix by the fraction of non-zero entries over  N^2
    if type(network) is not np.ndarray:
        raise (TypeError('Expect a np.ndarray as reservoir network'))

    # check if the matrix is square
    if network.shape[0] != network.shape[1]:    
        raise (ValueError('Expect network of square size!'))

    N = len(network)
    num_links = np.sum(network.flatten()>0)
    return num_links / (N**2)


def get_num_nodes(network: np.ndarray) -> int:

    # returns the number of nodes in the given network. Assumes the adjacency matrix to be of square size
    if type(network) is not np.ndarray:
        raise (TypeError('Expect a np.ndarray as reservoir network'))

    if network.shape[0] != network.shape[1]:
        raise (ValueError('Expect network of square size!'))

    non_zero_rows = np.any(network != 0, axis=1)

    # Identify columns that are not entirely zero
    non_zero_columns = np.any(network != 0, axis=0)

    # Identify nodes where both row and column are not zero
    non_isolated_nodes = np.where(non_zero_rows & non_zero_columns)[0]

    # Number of non-isolated nodes
    num_non_isolated_nodes = len(non_isolated_nodes)

    return num_non_isolated_nodes


def compute_spec_rad(network: np.ndarray) -> float:
    # compute the spectral radius of the network (max. eigenvalue)
    if type(network) is not np.ndarray:
        raise (TypeError('Expect a np.ndarray as reservoir network'))

    if network.shape[0] != network.shape[1]:
        raise (ValueError('Expect network of square size!'))

    return np.max(np.abs(np.linalg.eigvals(network)))


def set_spec_rad(network: np.ndarray, spec_radius:float) -> np.ndarray:
    # obtains a network with given spectral radius

    if spec_radius <= 0:
        raise(ValueError('spectral radius must be larger than zero'))
    elif (spec_radius > 1.0):
        raise(Warning('a spectral radius larger than 1 is unusual!'))

    # compute current spectral radius
    current_spectral_radius = compute_spec_rad(network)

    if current_spectral_radius < 10**(-9):
        print('spectral radius smaller than 10^-9!')
        current_spectral_radius = 10**(-6)

    scaling_factor = spec_radius / current_spectral_radius


    return network * scaling_factor


def is_zero_col_and_row(x: np.ndarray, idx: int) -> bool:
    # returns zero if adjacency matrix x carries only zeros in column and row of index idx (i.e. missing node)

    is_zero_column = np.all(x[:, idx] == 0)
    is_zero_row = np.all(x[idx, :] == 0)

    if is_zero_column and is_zero_row:
        return True
    else:
        return False


def remove_node(x: np.ndarray, idx: int | list) -> np.ndarray:
    if type(idx) is not list:
        idx = [idx]

    if x.ndim == 1:
        for idxx in idx:
            x[idxx] = 0
    elif x.ndim == 2:
        for idxx in idx:
            x[:, idxx] = 0
            x[idxx, :] = 0

    return x

def gen_init_states(num_nodes: int, method: str = 'random'):
    # returns an array of length <num_nodes>
    # creates the entries based on different sampling methods
    # when not setting specific values, the range is normalized to abs(1)

    if method == 'random':
        init_states = np.random.random(num_nodes)
    elif method == 'random_normal':
        init_states = np.random.randn(num_nodes)
    elif method == 'ones':
        init_states = np.ones(num_nodes)
    elif method == 'zeros':
        init_states = np.zeros(num_nodes)
    else:
        raise(ValueError(f'Sampling method {method} is unknown for generating initial reservoir states'))

    # normalize to max. absolute value of 1
    if method != 'zeros':
        init_states = init_states / np.max(np.abs(init_states))

    return init_states

def extract_density(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return nx.density(G)

def extract_spectral_radius(adjacency_matrix):
    return np.max(np.abs(np.linalg.eigvals(adjacency_matrix)))

def extract_in_degree_av(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    in_degrees = G.in_degree()
    avg_in_degree = np.mean(list(dict(in_degrees).values()))
    return avg_in_degree

def extract_out_degree_av(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    out_degrees = G.out_degree()
    avg_out_degree = np.mean(list(dict(out_degrees).values()))
    return avg_out_degree

def extract_clustering_coefficient(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return nx.average_clustering(G)

def extract_node_degree(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return dict(G.degree())

def extract_node_in_degree(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return dict(G.in_degree())

def extract_node_out_degree(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return dict(G.out_degree())

def extract_node_clustering_coefficient(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return nx.clustering(G)

def extract_node_betweenness_centrality(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return nx.betweenness_centrality(G)

def extract_node_pagerank(adjacency_matrix):
    G = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
    return nx.pagerank(G)

# Add more network property extraction functions as needed



"""
Some testing use case for the CustomModel API of pyreco
"""

from matplotlib import pyplot as plt
import sys
import os
import platform

from pyreco.custom_models import RC as RC
from pyreco.layers import InputLayer, ReadoutLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.plotting import r2_scatter
from pyreco.utils_data import scalar_to_scalar, vector_to_vector, sequence_to_sequence, x_to_x
# from sklearn.linear_model import Ridge
from pyreco.optimizers import RidgeSK

"""
Classic RC built on random networks. 

Use case: predict a sine signal into the future (n_time) samples from the past are mapped to (n_time) samples in the 
future
"""

# some testing data: predict a sine signal.
X_train, X_test, y_train, y_test = sequence_to_sequence(name='sine_pred', n_batch=20, n_states=2, n_time=200)

# set the dimensions (need to be [n_batch, n_time, n_states])
input_shape = (X_train.shape[1], X_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])

model_rc = RC()
model_rc.add(InputLayer(input_shape=input_shape))
model_rc.add(RandomReservoirLayer(nodes=100, density=0.1, activation='sigmoid', leakage_rate=0.1 , fraction_input=0.5))
model_rc.add(ReadoutLayer(output_shape, fraction_out=0.99))

# Compile the model
discard_transients = 100
optim = RidgeSK(alpha=0.5)
model_rc.compile(optimizer=optim, metrics=['mean_squared_error'], discard_transients=discard_transients)

# Train the model
model_rc.fit(X_train, y_train)

# Make predictions for new data
y_pred = model_rc.predict(X_test)
print(f'shape of predictions on test set: {y_pred.shape}')

# Evaluate the model
loss_rc = model_rc.evaluate(X_test, y_test, metrics=['mae'])
print(f'Test model loss: {loss_rc}')

# plot predictions vs. ground truth
r2_scatter(y_true=y_test[:,discard_transients:, :], y_pred=y_pred)

plt.figure()
plt.plot(y_test[0,discard_transients:,0], label='ground truth', marker='.')
plt.plot(y_pred[0,:,0], label='prediction', marker='.')
plt.title('test set')
plt.legend()
plt.xlabel('time')
plt.ylabel('x1')
plt.show()


"""
k-fold cross-validation of the model: estimate bias and variance of the model
"""
from pyreco.cross_validation import cross_val

val, mean, std_dev = cross_val(model_rc, X_train, y_train, n_splits=5, metric = ['mse'])
print(f"Cross-Validation MSE: {val}")
print(f"Mean MSE: {mean:.3f}")
print(f"Standard Deviation of MSE: {std_dev:.3f}")

